# -*- encoding: utf-8 -*-
from suite_py.lib import metrics
from suite_py.lib.handler.youtrack_handler import YoutrackHandler


class EstimateCone:
    def __init__(self, project, config, tokens):
        self._project = project
        self._config = config
        self._youtrack = YoutrackHandler(config, tokens)

    @metrics.command("estimate-cone")
    def run(self, issue: str, sprint_board: str, previous_sprints: int):
        remaining_story_points = self._sum_of_unresolved_descendant_story_points(issue)

        current_sprint = self._youtrack.get_current_sprint(sprint_board)
        sprints = self._youtrack.get_sprints(sprint_board)

        current_sprint_index = next(
            (
                index
                for index, sprint in enumerate(sprints)
                if sprint["id"] == current_sprint["id"]
            ),
            None,
        )

        recent_sprints = sprints[
            current_sprint_index - previous_sprints : current_sprint_index
        ]

        sprints_resolved_story_points = [
            self._youtrack.get_sprint_resolved_story_points(sprint_board, sprint["id"])
            for sprint in recent_sprints
        ]

        average_story_points_per_sprint = sum(sprints_resolved_story_points) / len(
            sprints_resolved_story_points
        )
        max_story_points_per_sprint = (
            max(sprints_resolved_story_points) or 1
        )  # Avoid division by zero
        min_story_points_per_sprint = (
            min(sprints_resolved_story_points) or 1
        )  # Avoid division by zero

        average_estimate = remaining_story_points / average_story_points_per_sprint
        best_estimate = remaining_story_points / max_story_points_per_sprint
        worst_estimate = remaining_story_points / min_story_points_per_sprint

        print(f"Remaining story points: {remaining_story_points}")

        print(f"Max story points per sprint: {max_story_points_per_sprint}")
        print(f"Average story points per sprint: {average_story_points_per_sprint:.2f}")
        print(f"Min story points per sprint: {min_story_points_per_sprint}")

        print(f"Average estimate: {average_estimate:.2f} sprints")
        print(f"Best estimate: {best_estimate:.2f} sprints")
        print(f"Worst estimate: {worst_estimate:.2f} sprints")

        print(f"Sprints resolved story points: {sprints_resolved_story_points}")

    def _sum_of_unresolved_descendant_story_points(self, _issue_readable_id):
        descendants = self._youtrack.get_issue_descendants(_issue_readable_id)

        descendants_digested = [
            {
                "id": descendant.get("idReadable"),
                "summary": descendant.get("summary"),
                "resolved": bool(descendant.get("resolved")),
                "story_points": next(
                    (
                        field["value"]
                        for field in descendant["customFields"]
                        if field["name"] == "Story Points"
                    )
                )
                or 0,
            }
            for descendant in descendants
        ]

        for descendant in descendants_digested:
            if not descendant["resolved"]:
                print(descendant)

        return sum(
            descendant["story_points"]
            for descendant in descendants_digested
            if not descendant["resolved"]
        )



#! -*- encoding: utf-8 -*-

# We import command modules inside their respective function bodies
# for performance reasons, so turn off the lint warning
# pylint: disable=import-outside-toplevel

# We need to inject truststore code before we do any libraries
# that might end up creating an ssl.SSlContext object
# pylint: disable=wrong-import-position

import sys


def maybe_inject_truststore() -> None:
    """
    Injects the truststore package into the system ssl store, to fix verficiation certificate issues when using warp.
    """
    if sys.version_info >= (3, 10):
        import truststore

        truststore.inject_into_ssl()
    else:
        # pylint: disable-next=reimported
        from suite_py.lib import logger

        logger.warning(
            "Your python version is older than 3.10 and doesn't support the truststore package. You might experience issues with certificate verification when connected to the warp VPN.\nPlease update to python 3.10 or newer"
        )


# Needs to be done before any other imports
maybe_inject_truststore()

import os
from functools import wraps
from importlib.metadata import PackageNotFoundError, version
from typing import Optional

import click
import requests
from click.exceptions import ClickException

from suite_py.__version__ import __version__
from suite_py.commands.context import Context
from suite_py.lib import logger, metrics
from suite_py.lib.config import Config
from suite_py.lib.handler import git_handler as git
from suite_py.lib.handler import prompt_utils
from suite_py.lib.handler.captainhook_handler import CaptainHook
from suite_py.lib.handler.okta_handler import Okta
from suite_py.lib.handler.pre_commit_handler import PreCommit
from suite_py.lib.tokens import Tokens

# pylint: enable=wrong-import-position

ALLOW_NO_GIT_SUBCOMMAND = ["check", "set-token", "login"]
ALLOW_NO_HOME_SUBCOMMAND = ["check", "set-token", "login"]


def fetch_latest_version() -> Optional[str]:
    """
    Fetches the latest version of suite-py as a str
    Returns None on error
    """
    try:
        # pylint: disable-next=missing-timeout
        pkg_info = requests.get("https://pypi.org/pypi/suite-py/json").json()
        return pkg_info["info"]["version"]
    except Exception:
        logger.warning("Failed to fetch latest version of suite-py!")
        return None


def upgrade_suite_py_if_needed(break_on_missing_package: bool = False) -> None:
    """
    Sometimes, when `suite-py` is launched with a virtual environment active, autoupgrade
    cannot see the package as installed (as the command is used from the "global" user environment)
    and raises an exception.
    """

    try:
        installed_version = version("suite_py")
    except PackageNotFoundError as error:
        # We are in a virtual environment where suite-py is not installed, don't bother trying to upgrade
        if break_on_missing_package:
            raise error

        logger.warning(
            "Skipping Suite-Py autoupgrade because the package was not found in the current environment."
        )
        return

    try:
        # If available, upgrade (if needed)
        latest_version = fetch_latest_version()
        if latest_version is None or latest_version > installed_version:
            from autoupgrade import Package

            # If we fail to fetch the latest version, fallback to attempting the upgrade anyway
            Package("suite_py").upgrade()
    except Exception as error:
        logger.warning(f"An error occurred whilst trying to upgrade suite-py: {error}")


# Catches any exceptions thrown and turns them into a ClickException
# So they don't print a stacktrace
def catch_exceptions(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            # Ignore all exception classes from the click module
            if type(e).__module__.startswith("click."):
                raise e

            logger.debug("An error occured:", exc_info=True)
            raise ClickException(f"{str(e)}. rerun with -v for more details") from e

    return wrapper


@click.group()
@click.option(
    "--project",
    type=click.Path(exists=True),
    default=os.getcwd(),
    help="Path of the project to run the command on (the default is current directory)",
)
@click.option(
    "--timeout",
    type=click.INT,
    help="Timeout in seconds for Captainhook operations",
)
@click.option("-v", "--verbose", count=True)
@click.pass_context
@catch_exceptions
def main(ctx, project, timeout, verbose):
    config = Config()

    logger.setup(verbose)

    logger.debug(f"v{__version__}")
    maybe_inject_truststore()
    upgrade_suite_py_if_needed(break_on_missing_package=False)

    if ctx.invoked_subcommand not in ALLOW_NO_GIT_SUBCOMMAND:
        project = git.get_root_folder(project)

    if ctx.invoked_subcommand not in ALLOW_NO_GIT_SUBCOMMAND and not git.is_repo(
        project
    ):
        raise ClickException(f"the folder {project} is not a git repo")

    if ctx.invoked_subcommand not in ALLOW_NO_HOME_SUBCOMMAND and not os.path.basename(
        project
    ) in os.listdir(config.user["projects_home"]):
        raise ClickException(
            f"the folder {project} is not in {config.user['projects_home']}"
        )

    skip_confirmation = False
    if type(config.user.get("skip_confirmation")).__name__ == "bool":
        skip_confirmation = config.user.get("skip_confirmation")
    elif type(
        config.user.get("skip_confirmation")
    ).__name__ == "list" and ctx.invoked_subcommand in config.user.get(
        "skip_confirmation"
    ):
        skip_confirmation = True

    if not skip_confirmation and not prompt_utils.ask_confirm(
        f"Do you want to continue on project {os.path.basename(project)}?"
    ):
        ctx.exit()
    if timeout:
        config.user["captainhook_timeout"] = timeout

    project = os.path.basename(project)
    tokens = Tokens()
    okta = Okta(config, tokens)
    captainhook = CaptainHook(config, okta, tokens)

    ctx.obj = Context(
        project=project,
        tokens=tokens,
        okta=okta,
        config=config,
        captainhook=captainhook,
    )

    ctx.obj.call(metrics.setup)

    # Skip chdir if not needed
    if (
        ctx.invoked_subcommand not in ALLOW_NO_GIT_SUBCOMMAND
        or ctx.invoked_subcommand not in ALLOW_NO_HOME_SUBCOMMAND
    ):
        os.chdir(os.path.join(config.user["projects_home"], ctx.obj.project))
    # Warn on missing pre_commit hook
    if ctx.invoked_subcommand not in ALLOW_NO_GIT_SUBCOMMAND:
        ctx.obj.call(PreCommit).check_and_warn()


@main.result_callback()
@click.pass_obj
def cleanup(_obj, _, **_kwargs):
    metrics.async_upload()


@main.command("bump", help="Bumps the project version based on the .versions.yml file")
@click.option("--project", required=False, type=str)
@click.option(
    "--version",
    required=False,
    type=str,
    help="Version to apply. If not passed, you will be prompted to insert or select one from a predefined list",
)
@click.pass_obj
@catch_exceptions
def bump(obj: Context, project: Optional[str] = None, version: Optional[str] = None):
    from suite_py.commands.bump import Bump

    obj.call(Bump).run(project=project, version=version)


@main.command(
    "create-branch", help="Create local branch and set the YouTrack card in progress"
)
@click.option("--card", type=click.STRING, help="YouTrack card number (ex. PRIMA-123)")
@click.option(
    "--autostash",
    is_flag=True,
    help="Stash uncommitted changes before creating the branch and reapply them afterward",
)
@click.pass_obj
@catch_exceptions
def cli_create_branch(obj, card, autostash):
    from suite_py.commands.create_branch import CreateBranch

    obj.call(CreateBranch, card=card, autostash=autostash).run()


@main.command("lock", help="Lock project on staging or prod")
@click.argument(
    "environment", type=click.Choice(("staging", "production", "deploy", "merge"))
)
@click.pass_obj
@catch_exceptions
def cli_lock_project(obj, environment):
    from suite_py.commands.project_lock import ProjectLock

    obj.call(ProjectLock, env=environment, action="lock").run()


@main.command("unlock", help="Unlock project on staging or prod")
@click.argument(
    "environment", type=click.Choice(("staging", "production", "deploy", "merge"))
)
@click.pass_obj
@catch_exceptions
def cli_unlock_project(obj, environment):
    from suite_py.commands.project_lock import ProjectLock

    obj.call(ProjectLock, env=environment, action="unlock").run()


@main.command(
    "estimate-cone",
    help="Point-in-time estimate of the time needed to complete all unresolved children of a card using the Cone of Uncertainty",
)
@click.option("--issue", prompt="Issue ID", type=str)
@click.option("--sprint-board", prompt="Sprint Board ID", type=str)
@click.option("--previous-sprints", required=False, type=int, default=6)
@click.pass_obj
@catch_exceptions
def estimate_cone(
    obj: Context,
    issue: Optional[str],
    sprint_board: Optional[str],
    previous_sprints: int,
):
    from suite_py.commands.estimate_cone import EstimateCone

    obj.call(EstimateCone).run(
        issue=issue, sprint_board=sprint_board, previous_sprints=previous_sprints
    )


@main.command("open-pr", help="Open a PR on GitHub")
@click.pass_obj
@catch_exceptions
def cli_open_pr(obj):
    from suite_py.commands.ask_review import AskReview
    from suite_py.commands.open_pr import OpenPR

    ask_review = obj.call(AskReview)
    obj.call(OpenPR, ask_review=ask_review).run()


@main.command("ask-review", help="Requests a PR review")
@click.pass_obj
@catch_exceptions
def cli_ask_review(obj):
    from suite_py.commands.ask_review import AskReview

    obj.call(AskReview).run()


@main.command(
    "merge-pr", help="Merge the selected branch to master if all checks are OK"
)
@click.pass_obj
@catch_exceptions
def cli_merge_pr(obj):
    from suite_py.commands.merge_pr import MergePR

    obj.call(MergePR).run()


@main.group("release", help="Manage releases")
def release():
    pass


@release.command(
    "create", help="Create a github release (and deploy it if GHA are used)"
)
@click.argument(
    "commit",
    required=False,
    type=str,
    metavar="[COMMIT]",
)
@click.option(
    "--interactive",
    "-i",
    is_flag=True,
    help="Interactively choose which unreleased commit to release (ignored if COMMIT is provided)",
)
@click.pass_obj
@catch_exceptions
def cli_release_create(obj, commit, interactive):  # type: ignore[override]
    """Create a release.

    Optionally pass a COMMIT (full or short SHA) to create the release at that
    specific commit instead of the latest commit (HEAD). Example:

        suite-py release create 644a699

    If you don't provide a COMMIT you can pass --interactive / -i to select
    one among the unreleased commits (those after the last tag).
    """
    from suite_py.commands.release import Release

    obj.call(Release, action="create", commit=commit, interactive=interactive).run()


@main.command("status", help="Current status of a project")
@click.pass_obj
@catch_exceptions
def cli_status(obj):
    from suite_py.commands.status import Status

    obj.call(Status).run()


@main.command("check", help="Verify authorisations for third party services")
@click.pass_obj
@catch_exceptions
def cli_check(obj):
    from suite_py.commands.check import Check

    obj.call(Check).run()


@main.command("login", help="Manage login against Okta")
@click.pass_obj
@catch_exceptions
def login(obj):
    from suite_py.commands.login import Login

    obj.call(Login).run()


@main.command("set-token", help="Create or update a service token")
@click.pass_obj
@catch_exceptions
def cli_set_token(obj):
    from suite_py.commands.set_token import SetToken

    obj.call(SetToken).run()



import base64
import hashlib
import secrets
import typing
import urllib.parse
import webbrowser
from dataclasses import dataclass
from http.server import BaseHTTPRequestHandler, HTTPServer
from os import path
from urllib.parse import parse_qs, urlparse

import requests

import suite_py
from suite_py.lib import logger


@dataclass
class OAuthTokenResponse:
    access_token: str
    id_token: typing.Optional[str]
    refresh_token: typing.Optional[str]
    expires_in: float


class OAuthError(Exception):
    pass


def retrieve_token(base_url, params) -> OAuthTokenResponse:
    url = f"{base_url}/token"
    headers = {
        "content-type": "application/x-www-form-urlencoded",
        "Accept": "application/json",
    }

    data = requests.post(url, headers=headers, data=params, timeout=30).json()
    logger.debug(data)

    if error := data.get("error_description", data.get("errorSummary", None)):
        raise OAuthError(f"OAuth error: {error}")

    return OAuthTokenResponse(
        access_token=data["access_token"],
        expires_in=data["expires_in"],
        id_token=data.get("id_token", None),
        refresh_token=data.get("refresh_token", None),
    )


class OAuthCallbackServer(HTTPServer):
    received_state: typing.Optional[str] = None
    error_message: typing.Optional[str] = None
    code: typing.Optional[str] = None

    def __init__(self, server_address) -> None:
        super().__init__(server_address, OAuthCallbackRequestHandler)


class OAuthCallbackRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        assert isinstance(self.server, OAuthCallbackServer)

        server = self.server
        args = parse_qs(urlparse(self.path).query)

        server.received_state = args["state"][0]
        if "error" in args:
            error = args["error"][0]
            error_description = args["error_description"][0]
            server.error_message = f"{error}: {error_description}"
        else:
            server.code = args["code"][0]

        self.send_response(200)
        self.send_header("content-type", "text/html")
        self.end_headers()

        template = path.join(path.dirname(suite_py.__file__), "templates/login.html")
        with open(template, "rb") as f:
            self.wfile.write(f.read())


def _url_encode_no_padding(byte_data):
    """
    Safe encoding handles + and /, and also replace = with nothing
    """
    return base64.urlsafe_b64encode(byte_data).decode("utf-8").replace("=", "")


def _generate_challenge(a_verifier):
    return _url_encode_no_padding(hashlib.sha256(a_verifier.encode()).digest())


def authorization_code_flow(
    client_id,
    base_url,
    scope,
    redirect_uri="http://127.0.0.1:5000/callback",
    listen=("127.0.0.1", 5000),
):
    # From https://developer.okta.com/docs/guides/sign-into-web-app-redirect/python/main/
    # Step1: Create code verifier: Generate a code_verifier that will be sent to Okta to request tokens.
    verifier = _url_encode_no_padding(secrets.token_bytes(32))
    # Step2: Create code challenge: Generate a code_challenge from the code_verifier that will be sent to Okta to request an authorization_code.
    challenge = _generate_challenge(verifier)
    state = _url_encode_no_padding(secrets.token_bytes(32))

    # We generate a nonce (state) that is used to protect against attackers invoking the callback
    url = f"{base_url}/authorize?"
    url_parameters = {
        "scope": scope,
        "response_type": "code",
        "redirect_uri": redirect_uri,
        "client_id": client_id,
        "code_challenge": challenge.replace("=", ""),
        "code_challenge_method": "S256",
        "state": state,
    }
    url = url + urllib.parse.urlencode(url_parameters)

    # Step3: Authorize user: Request the user's authorization and redirect back to your app with an authorization_code.
    # Open the browser window to the login url
    # Start the server
    logger.info("A browser tab should've opened. If not manually navigate to: " + url)
    webbrowser.open(url)

    server = OAuthCallbackServer(listen)
    server.handle_request()

    if state != server.received_state:
        raise OAuthError(
            "Error: session replay or similar attack in progress. Please log out of all connections."
        )

    if server.error_message:
        raise OAuthError(server.error_message)

    # Step4: Request tokens: Exchange your authorization_code and code_verifier for tokens.
    body = {
        "grant_type": "authorization_code",
        "client_id": client_id,
        "code_verifier": verifier,
        "code": server.code,
        "redirect_uri": redirect_uri,
    }
    return retrieve_token(base_url, body)


def do_refresh_token(
    client_id: str, base_url: str, scope: str, refresh_token: str
) -> OAuthTokenResponse:
    # See https://developer.okta.com/docs/guides/refresh-tokens/main/
    params = {
        "scope": scope,
        "client_id": client_id,
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
    }
    return retrieve_token(base_url, params)



import dataclasses
from inspect import signature

from suite_py.lib.config import Config
from suite_py.lib.handler.captainhook_handler import CaptainHook
from suite_py.lib.handler.okta_handler import Okta
from suite_py.lib.tokens import Tokens


@dataclasses.dataclass
class Context:
    project: str
    config: Config
    captainhook: CaptainHook
    tokens: Tokens
    okta: Okta

    # Call the function to_call with kwargs, injecting fields from self as default arguments
    def call(self, to_call, **kwargs):
        provided = self.shallow_dict()
        needed = signature(to_call).parameters.keys()
        provided = {k: provided[k] for k in needed if k in provided}

        kwargs = provided | kwargs

        return to_call(**kwargs)

    def shallow_dict(self):
        """
        Converts the dataclass to a dict.

        Unlike dataclasses.asdict this function only shallow copies the fields
        instead of using copy.deepcopy()
        """
        return {
            field.name: getattr(self, field.name) for field in dataclasses.fields(self)
        }



# -*- encoding: utf-8 -*-
import logging

from rich.logging import RichHandler

# Rich handles color and formatting internally
# This format will make it so we print stuff like "[I] message"
LOG_FORMAT = "%(message)s"

LOGGER_NAME = "suite_py_logger"


def setup(verbose):
    level = logging.DEBUG if verbose else logging.INFO

    logging.basicConfig(
        level=level,
        format=LOG_FORMAT,
        datefmt="[%X]",
        handlers=[
            RichHandler(
                rich_tracebacks=True,
                markup=False,
                show_path=False,
                show_level=False,
                show_time=False,
            )
        ],
    )

    logger = logging.getLogger(LOGGER_NAME)
    logger.debug("Logging as %s", level)


def debug(message, *args, **kwargs):
    logging.getLogger(LOGGER_NAME).debug(message, *args, **kwargs)


def info(message, *args, **kwargs):
    logging.getLogger(LOGGER_NAME).info(message, *args, **kwargs)


def warning(message, *args, **kwargs):
    logging.getLogger(LOGGER_NAME).warning(message, *args, **kwargs)


def error(message, *args, **kwargs):
    logging.getLogger(LOGGER_NAME).error(message, *args, **kwargs)



# -*- coding: utf-8 -*-
import re
import sys

from suite_py.commands import common
from suite_py.lib import logger, metrics
from suite_py.lib.handler import git_handler as git
from suite_py.lib.handler import prompt_utils
from suite_py.lib.handler.changelog_handler import ChangelogHandler
from suite_py.lib.handler.git_handler import GitHandler
from suite_py.lib.handler.github_handler import GithubHandler
from suite_py.lib.handler.version_handler import DEFAULT_VERSION, VersionHandler
from suite_py.lib.handler.youtrack_handler import YoutrackHandler


class Release:
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-positional-arguments
    def __init__(
        self,
        action,
        project,
        captainhook,
        config,
        tokens,
        commit=None,
        interactive=False,
    ):
        self._action = action
        self._project = project
        self._config = config
        self._tokens = tokens
        # Optional commit (sha or short sha) to release instead of HEAD
        self._commit = commit
        # Whether to prompt the user to choose commit among unreleased ones
        self._interactive = interactive
        self._changelog_handler = ChangelogHandler()
        self._youtrack = YoutrackHandler(config, tokens)
        self._captainhook = captainhook
        self._github = GithubHandler(tokens)
        self._repo = self._github.get_repo(project)
        self._git = GitHandler(project, config)
        self._version = VersionHandler(self._repo, self._git, self._github)

    @metrics.command("release")
    def run(self):
        self._stop_if_prod_locked()
        self._git.fetch()

        if self._action == "create":
            self._create()

    def _stop_if_prod_locked(self):
        request = self._captainhook.status(self._project, "production")
        if request.status_code != 200:
            logger.error("Unable to determine lock status on master.")
            sys.exit(-1)

        request_object = request.json()
        if request_object["locked"]:
            logger.error(
                f"The project is locked in production by {request_object['by']}. Unable to continue."
            )
            sys.exit(-1)

    def _create(self):
        latest = self._version.get_latest_version()
        commits, new_version, message = self._gather_commits_and_version(latest)
        message = self._augment_message_with_changelog(new_version, message)
        message = common.ask_for_release_description(message)
        sha = self._resolve_target_sha(commits)
        self._create_release(new_version, message, sha)

    # ---- helpers (extracted to reduce complexity of _create) ----
    def _gather_commits_and_version(self, latest):
        """Return (commits, new_version, base_message)."""
        if latest == "":
            return self._first_release_flow()

        logger.info(f"The current release is {latest}")
        commits = self._github.get_commits_since_release(self._repo, latest)
        # If user did not pass a target commit, ask interactively which commit to release.
        # Only perform interactive selection if user explicitly requested it
        if self._interactive:
            self._select_commit_if_needed(commits)
        commits = self._maybe_trim_commits(commits)
        _check_migrations_deploy(commits)
        message = self._build_commits_message(commits)
        logger.info(f"\nCommits list:\n{message}\n")
        if not prompt_utils.ask_confirm("Do you want to continue?"):
            sys.exit()  # Preserve previous early return semantics
        new_version = self._version.select_new_version(latest, allow_prerelease=True)
        return commits, new_version, message

    def _first_release_flow(self):
        logger.warning(f"No tags found, I'm about to push the tag {DEFAULT_VERSION}")
        if not prompt_utils.ask_confirm(
            "Are you sure you want to continue?", default=False
        ):
            sys.exit()
        return [], DEFAULT_VERSION, f"First release with tag {DEFAULT_VERSION}"

    def _maybe_trim_commits(self, commits):
        if not self._commit:
            return commits
        try:
            resolved_commit = self._repo.get_commit(self._commit)
            target_sha = resolved_commit.sha
        except Exception:  # pragma: no cover
            logger.error(
                f"The provided commit '{self._commit}' was not found in the repository."
            )
            sys.exit(-1)
        target_index = None
        for idx, c in enumerate(commits):
            if (
                c.sha == target_sha
                or getattr(c, "commit", None)
                and c.commit.sha == target_sha
            ):
                target_index = idx
                break
        if target_index is None:
            logger.error(
                "The specified commit is not part of the unreleased commits (i.e., not after the latest release)."
            )
            sys.exit(-1)
        return commits[target_index:]

    def _select_commit_if_needed(self, commits):
        """Interactively ask the user which commit to release if none was provided.

        Presents the unreleased commits (newest first) allowing the user to pick a
        target commit. The chosen commit's SHA (full) is stored in `self._commit` so
        that downstream logic (_maybe_trim_commits / _resolve_target_sha) works
        unchanged.
        """
        if self._commit or not commits:
            return
        if len(commits) == 1:
            # Only one unreleased commit; implicitly choose it.
            self._commit = commits[0].sha
            return
        choices = [
            {
                "name": f"{c.sha[:8]} | {c.commit.message.splitlines()[0]}",
                "value": c.sha,
            }
            for c in commits
        ]
        selected = prompt_utils.ask_choices(
            "Select commit to release (newest first):", choices
        )
        self._commit = selected

    @staticmethod
    def _build_commits_message(commits):
        return "\n".join(
            [
                "* " + c.commit.message.splitlines()[0] + " by " + c.commit.author.name
                for c in commits
            ]
        )

    def _augment_message_with_changelog(self, new_version, message):
        if not self._changelog_handler.changelog_exists():
            return message
        latest_tag, latest_entry = self._changelog_handler.get_latest_entry_with_tag()
        if latest_tag != new_version:
            if not prompt_utils.ask_confirm(
                "You didn't update your changelog, are you sure you want to proceed?"
            ):
                sys.exit()
            return message
        return f"{latest_entry}\n\n# Commits\n\n{message}"

    def _resolve_target_sha(self, commits):
        if not self._commit:
            return commits[0].commit.sha if commits else ""
        try:
            sha = self._repo.get_commit(self._commit).sha
        except Exception:  # pragma: no cover
            logger.error(
                f"Unable to resolve commit '{self._commit}' during release creation."
            )
            sys.exit(-1)
        self._validate_target_commit_not_tagged(sha)
        return sha

    def _validate_target_commit_not_tagged(self, sha):
        try:
            for t in self._repo.get_tags():  # network call
                if t.commit.sha == sha:
                    logger.error(
                        f"The commit {sha[:7]} is already tagged with {t.name}. Aborting."
                    )
                    sys.exit(-1)
        except Exception:  # pragma: no cover
            logger.warning("Could not verify if commit already tagged; continuing.")

    def _create_release(self, new_version, message, commit):
        new_release = self._repo.create_git_release(
            new_version,
            new_version,
            self._youtrack.replace_card_names_with_md_links(message),
            target_commitish=commit,
        )
        if new_release:
            logger.info(f"The release has been created! Link: {new_release.html_url}")

    def _manage_youtrack_card(self, version, countries):
        release_state = self._config.youtrack["release_state"]

        release_body = self._repo.get_release(version).body

        issue_ids = self._youtrack.get_ids_from_release_body(release_body)

        if len(issue_ids) > 0:
            update_youtrack_state = prompt_utils.ask_confirm(
                f"Do you want to move the associated cards to {release_state} state?",
                default=False,
            )

            for issue_id in issue_ids:
                try:
                    self._youtrack.comment(
                        issue_id,
                        f"Deploy in production of {self._project} in countries {countries} done with the release {version}",
                    )
                    if update_youtrack_state:
                        self._youtrack.update_state(issue_id, release_state)
                        logger.info(f"{issue_id} moved to {release_state}")
                except Exception:
                    logger.warning(
                        f"An error occurred while moving the card {issue_id} to {release_state}"
                    )
                repos_status = self._get_repos_status_from_issue(issue_id)
                if all(r["deployed"] for r in repos_status.values()):
                    try:
                        self._youtrack.update_deployed_field(issue_id)
                        logger.info("Custom field Deployed updated on YouTrack")
                    except Exception:
                        logger.warning(
                            "An error occurred while updating the custom field Deployed"
                        )

    def _get_repos_status_from_issue(self, issue_id):
        regex_pr = r"^PR .* -> https:\/\/github\.com\/primait\/(.*)\/pull\/([0-9]*)$"
        regex_deploy = r"^Deploy in production of (.*) in countries"
        comments = self._youtrack.get_comments(issue_id)
        repos_status = {}

        for c in comments:
            m = re.match(regex_pr, c["text"])
            if m:
                project = m.group(1)
                pr_number = int(m.group(2))
                repos_status[project] = {}
                repos_status[project]["pr"] = pr_number
                repos_status[project]["deployed"] = False
            m = re.match(regex_deploy, c["text"])
            if m:
                project = m.group(1)
                try:
                    repos_status[project]["deployed"] = True
                except Exception:
                    pass
        return repos_status

    def _tags_drifted(self, versions):
        for country, version in versions.items():
            for c, v in versions.items():
                if country == c:
                    continue
                if version is None or v is None or version.compare(v) != 0:
                    return True
        return False


def _check_migrations_deploy(commits):
    if not commits:
        logger.error("ERROR: no commit found")
        sys.exit(-1)
    elif len(commits) == 1:
        files_changed = git.files_changed_between_commits("--raw", f"{commits[0].sha}~")
    else:
        files_changed = git.files_changed_between_commits(
            f"{commits[-1].sha}~", commits[0].sha
        )
    if git.migrations_found(files_changed):
        logger.warning("WARNING: migrations detected in the code")
        if not prompt_utils.ask_confirm(
            "Are you sure you want to continue?", default=False
        ):
            sys.exit()



# -*- coding: utf-8 -*-

from suite_py.lib import logger, metrics
from suite_py.lib.handler import prompt_utils
from suite_py.lib.handler.github_handler import GithubHandler
from suite_py.lib.handler.youtrack_handler import YoutrackHandler
from suite_py.lib.symbol import CHECKMARK, CROSSMARK


class Check:
    def __init__(self, captainhook, config, tokens):
        self.config = config
        self._invalid_or_missing_tokens = []
        self._tokens = tokens
        self._youtrack = YoutrackHandler(config, tokens)
        self._github = GithubHandler(tokens)
        self._captainhook = captainhook

        self._checks = [
            ("Github", self._check_github),
            ("Youtrack", self._check_youtrack),
            ("CaptainHook", self._check_captainhook),
            ("AWS", self._check_aws),
        ]

    @metrics.command("check")
    def run(self):
        # Services
        for service, check in self._checks:
            _forge_message(service, check())

        # Tokens
        if len(self._invalid_or_missing_tokens) > 0 and prompt_utils.ask_confirm(
            "Do you want to reinsert the missing tokens?", default=True
        ):
            self._refresh_tokens()

    def _check_github(self):
        if self._tokens.github:
            try:
                if self._github.get_organization():
                    return "ok"
                self._invalid_or_missing_tokens.append("github")
                return "invalid_token"
            except Exception:
                self._invalid_or_missing_tokens.append("github")
                return "invalid_token"
        else:
            self._invalid_or_missing_tokens.append("github")
            return "missing_token"

    def _check_youtrack(self):
        if self._tokens.youtrack:
            try:
                self._youtrack.get_projects()
                return "ok"
            except Exception:
                self._invalid_or_missing_tokens.append("youtrack")
                return "invalid_token"
        else:
            self._invalid_or_missing_tokens.append("youtrack")
            return "missing_token"

    def _check_captainhook(self):
        try:
            if self._captainhook.check().status_code != 200:
                return "unreachable"
            return "ok"
        except Exception:
            return "unreachable"

    def _check_aws(self):
        try:
            if self.config.aws != "":
                return "ok"
            return "missing_token"
        except Exception:
            return "missing_token"

    def _refresh_tokens(self):
        for service in self._invalid_or_missing_tokens:
            new_token = prompt_utils.ask_questions_input(
                f"Enter new token for {service}: "
            )
            self._tokens.edit(service, new_token)

        logger.info("Saving the new tokens ...")
        self._tokens.save()
        logger.info(f"{CHECKMARK} Done!")


def _forge_message(service, result):
    cases = {
        "ok": f"{service:>12}:{CHECKMARK:>12} ok",
        "invalid_token": f"{service:>12}:{CROSSMARK:>12} token not valid",
        "missing_token": f"{service:>12}:{CROSSMARK:>12} missing token",
        "unreachable": f"{service:>12}:{CROSSMARK:>12} unreachable",
    }
    print(cases.get(result, f"{service}: unknown state"))



import os

import yaml

from suite_py.lib import logger
from suite_py.lib.config import Config
from suite_py.lib.handler.git_handler import GitHandler


class PreCommit:
    """
    Handler checking whether the user has the prima pre-commit hooks setup and print a warning if they are missing
    """

    def __init__(self, project: str, config: Config):
        self._git = GitHandler(project, config)

    def check_and_warn(self):
        if self._is_enabled() and not self._is_pre_commit_hooks_installed():
            self._warn_missing_pre_commit_hook()

    def _is_enabled(self):
        return self._git.get_git_config("suite-py.disable-pre-commit-warning") != "true"

    def _is_pre_commit_hooks_installed(self):
        """
        Apply some heuristics to check whether the gitleaks pre-commit hook is installed.
        This is extremely imperfect, and only supports direct calls in shell scripts.
        """
        return self._is_shell_script_hook_setup() or self._is_pre_commit_py_hook_setup()

    def _warn_missing_pre_commit_hook(self):
        logger.warning(
            """
Looks like the current repo is missing the gitleaks pre-commit hook!
Please install it per the security guide:
https://www.notion.so/helloprima/Install-Gitleaks-pre-commit-hook-aaaa6beafafa4c298b537afcb52bb25a

If you have installed them already you can report the false positive to team-platform-shared-services (on Slack) and run:
    git config suite-py.disable-pre-commit-warning true
to disable the check for this repo, or
    git config --global suite-py.disable-pre-commit-warning true
to disable it globally
        """
        )

    def _is_shell_script_hook_setup(self):
        """
        Check whether the gitleaks hook is setup as a regular bash script:

        * is there a `.git/hooks/pre-commit` shell script that contains keyword "gitleaks"
        * is there a `.husky/pre-commit` shell script that contains keyword "security-hooks" (primait/security-hooks repo)
        """
        checks = [
            (os.path.join(self._git.hooks_path(), "pre-commit"), "gitleaks"),
            (
                os.path.join(self._git.get_path(), ".husky", "pre-commit"),
                "security-hooks",
            ),
        ]

        return any(
            self._script_contains_keyword(path, keyword) for path, keyword in checks
        )

    def _is_pre_commit_py_hook_setup(self):
        """
        Check whether the gitleaks hook is setup with the pre-commit python framework
        """
        # is there a `.git/hooks/pre-commit` shell script that contains keyword "pre-commit"
        pre_commit_file_path = os.path.join(self._git.hooks_path(), "pre-commit")
        if not self._script_contains_keyword(pre_commit_file_path, "pre-commit"):
            logger.debug("pre-commit.com not installed, skipping config check")
            return False

        config_path = os.path.join(self._git.get_path(), ".pre-commit-config.yaml")
        try:
            with open(config_path, encoding="utf-8") as f:
                config = yaml.safe_load(f)
        except FileNotFoundError:
            logger.debug("pre-commit config file(%s) not found", config_path)
            return False
        except yaml.YAMLError:
            logger.warning(".pre-commit-config.yaml file is invalid!", exc_info=True)
            return False

        return any(
            repo.get("repo", "") == "git@github.com:primait/security-hooks.git"
            for repo in config.get("repos", [])
        )

    def _script_contains_keyword(self, file_path, keyword):
        """
        Check if a keyword appears in a shell script, ignoring comment lines
        (binaries and python code are out of scope for us).
        """
        try:
            logger.debug("checking pre-commit script(%s)", file_path)
            with open(file_path, encoding="utf-8") as f:
                content = f.read()
        except FileNotFoundError:
            logger.debug("pre-commit script(%s) not found", file_path)
            return False

        # Filter out comments (lines starting with '#').
        lines_without_comments = (
            line for line in content.splitlines() if not line.strip().startswith("#")
        )
        return any(keyword in line for line in lines_without_comments)



# -*- encoding: utf-8 -*-
import requests

from suite_py.__version__ import __version__
from suite_py.lib.handler.github_handler import GithubHandler
from suite_py.lib.handler.okta_handler import Okta


class UnauthorizedError(Exception):
    def __init__(self) -> None:
        super().__init__("Unauthorized with captainhook")


class CaptainHook:
    _okta: Okta
    _github: GithubHandler

    def __init__(self, config, okta: Okta, tokens=None):
        self._baseurl = config.user["captainhook_url"]
        self._timeout = config.user["captainhook_timeout"]
        self._okta = okta

        if tokens is not None:
            self._github = GithubHandler(tokens)

    def lock_project(self, project, env):
        data = {
            "project": project,
            "status": "locked",
            "user": self._get_user(),
            "environment": env,
        }
        return self.send_post_request("/projects/manage-lock", data)

    def unlock_project(self, project, env):
        data = {
            "project": project,
            "status": "unlocked",
            "user": self._get_user(),
            "environment": env,
        }
        return self.send_post_request("/projects/manage-lock", data)

    def status(self, project, env):
        return self.send_get_request(
            f"/projects/check?project={project}&environment={env}"
        )

    def check(self):
        return requests.get(f"{self._baseurl}/", timeout=(2, self._timeout))

    def get_users_list(self):
        return self.send_get_request("/users/all")

    def send_metrics(self, metrics):
        self.send_post_request("/suite_py/metrics/", json=metrics).raise_for_status()

    def send_post_request(self, endpoint, data=None, json=None):
        r = requests.post(
            f"{self._baseurl}{endpoint}",
            headers=self._headers(),
            data=data,
            json=json,
            timeout=self._timeout,
        )

        return self._response(r)

    def send_put_request(self, endpoint, data=None, json=None):
        r = requests.put(
            f"{self._baseurl}{endpoint}",
            headers=self._headers(),
            data=data,
            json=json,
            timeout=self._timeout,
        )

        return self._response(r)

    def send_get_request(self, endpoint):
        r = requests.get(
            f"{self._baseurl}{endpoint}",
            headers=self._headers(),
            timeout=(2, self._timeout),
        )

        return self._response(r)

    def _response(self, response):
        if response.status_code == 401:
            raise UnauthorizedError()

        return response

    def set_timeout(self, timeout):
        self._timeout = timeout

    def _get_user(self):
        return self._github.get_user().login

    def _headers(self):
        id_token = self._okta.get_id_token()
        return {
            "User-Agent": f"suite-py/{__version__}",
            "Authorization": f"Bearer {id_token}",
        }
